﻿// Copyright 2012-2014 Andrew C. Dvorak
//
// This file is part of BDHero.
//
// BDHero is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// BDHero is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with BDHero.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Security;
using System.Windows.Forms;
using System.Windows.Forms.Design;
using BDHero.Utils;
using DotNetUtils.Annotations;
using DotNetUtils.Concurrency;
using OSUtils.DriveDetector;
using UILib.WinForms;

namespace BDHeroGUI.Components
{
    [DefaultProperty("NoDiscText")]
    [DefaultEvent("DiscSelected")]
    [ToolStripItemDesignerAvailability(ToolStripItemDesignerAvailability.StatusStrip)]
    [ToolboxItem(true)]
    public partial class DiscMenu : ToolStripMenuItem
    {
        private static readonly log4net.ILog Logger =
            log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        #region Default text values

        private const string DefaultNoDiscText = "No discs found";
        private const string DefaultScanningText = "Scanning for discs...";

        #endregion

        #region Public members

        [Category("Appearance")]
        [Description("The text that is displayed when no discs are found.")]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
        [Localizable(true)]
        [DefaultValue(DefaultNoDiscText)]
        [Browsable(true)]
        [EditorBrowsable(EditorBrowsableState.Always)]
        public string NoDiscText
        {
            get { return _noDiscItem.Text; }
            set { _noDiscItem.Text = value; }
        }

        [Category("Appearance")]
        [Description("The text that is displayed while the menu is scanning for discs.")]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
        [Localizable(true)]
        [DefaultValue(DefaultScanningText)]
        [Browsable(true)]
        [EditorBrowsable(EditorBrowsableState.Always)]
        public string ScanningText
        {
            get { return _scanningItem.Text; }
            set { _scanningItem.Text = value; }
        }

        [Description("Invoked whenever a disc is selected (clicked on) by the user.")]
        [Browsable(true)]
        [EditorBrowsable(EditorBrowsableState.Always)]
        public event DiscMenuItemClickHandler DiscSelected;

        #endregion

        /// <summary>
        /// Gets a collection of all BD-ROM drives connected to the host OS.
        /// </summary>
        private static DriveInfo[] Drives
        {
            get { return DriveInfo.GetDrives().Where(BDFileUtils.IsBDROM).ToArray(); }
        }

        /// <summary>
        /// Gets ALL menu items present in the dropdown list.
        /// </summary>
        private ToolStripItem[] AllMenuItems
        {
            get { return DropDownItems.OfType<ToolStripItem>().ToArray(); }
        }

        private readonly ToolStripMenuItem _dummyItem = new ToolStripMenuItem("DUMMY") { Enabled = false };
        private readonly ToolStripMenuItem _noDiscItem = new ToolStripMenuItem(DefaultNoDiscText) { Enabled = false };
        private readonly ToolStripMenuItem _scanningItem = new ToolStripMenuItem(DefaultScanningText) { Enabled = false };
        private readonly ToolStripSeparator _dividerItem = new ToolStripSeparator();

        private IDriveDetector _detector;
        private bool _isScanning;
        private bool _isInitialized;

        #region Constructors

        public DiscMenu()
        {
            InitializeComponent();
        }

        public DiscMenu(IContainer container)
        {
            container.Add(this);

            InitializeComponent();
        }

        #endregion

        #region Initialization

        /// <summary>
        /// Initializes the <see cref="BDHeroGUI.Components.DiscMenu"/> for use.
        /// </summary>
        /// <param name="form">
        /// Windows Forms control (typically a <see cref="Form"/>) to listen for <see cref="Form.WndProc"/> events on.
        /// </param>
        /// <param name="detector">
        /// Drive detector.
        /// </param>
        /// <exception cref="InvalidOperationException">Thrown if this method is called more than once.</exception>
        public void Initialize(Form form, IDriveDetector detector)
        {
            if (_isInitialized)
            {
                throw new InvalidOperationException("DiscMenu has already been initialized");
            }

            var hook = new WndProcHook(form);

            hook.WndProcMessage += WndProc;

            _detector = detector;
            _detector.DeviceArrived += OnDeviceArrived;
            _detector.DeviceRemoved += OnDeviceRemoved;

            DropDownOpened += OnDropDownOpened;
            DropDownClosed += OnDropDownClosed;

            Refresh();

            _isInitialized = true;
        }

        #endregion

        #region Event handlers

        private void OnDeviceArrived(object sender, DriveDetectorEventArgs args)
        {
            Logger.InfoFormat("Device arrived: {0}", args.DriveInfo);
            Refresh();
        }

        private void OnDeviceRemoved(object sender, DriveDetectorEventArgs args)
        {
            Logger.InfoFormat("Device removed: {0}", args.DriveInfo);
            Refresh();
        }

        private void OnDropDownOpened(object sender, EventArgs eventArgs)
        {
            Logger.Debug("DropDown opened");
            Refresh();
        }

        private void OnDropDownClosed(object sender, EventArgs eventArgs)
        {
            Logger.Debug("DropDown closed");
        }

        private void WndProc(ref Message m, HandledEventArgs args)
        {
            _detector.WndProc(ref m);
        }

        private void MenuItemOnClick(object sender, EventArgs eventArgs)
        {
            var menuItem = sender as ToolStripMenuItem;
            if (menuItem == null) return;

            var driveInfo = menuItem.Tag as DriveInfo;
            if (driveInfo == null) return;

            if (DiscSelected != null)
            {
                DiscSelected(driveInfo);
            }
        }

        #endregion

        #region Disc scanning and menu population logic

        private void Refresh()
        {
            ScanAsync();
        }

        private void ScanAsync()
        {
            if (_isScanning)
            {
                Logger.Info("Already scanning for discs; ignoring");
                return;
            }

            Logger.Info("Scanning for discs...");

            _isScanning = true;

            DropDownItems.Add(_dividerItem);
            DropDownItems.Add(_scanningItem);

            var menuItems = new ToolStripItem[0];

            new EmptyPromise(Parent)
                .Work(promise => menuItems = CreateToolStripItems(Drives))
                .Done(promise => UpdateMenu(menuItems))
                .Fail(promise => Logger.Error("Error occurred while scanning for discs", promise.LastException))
                .Always(promise => _isScanning = false)
                .Start();
        }

        private void ClearMenu()
        {
            // We need to always keep at least 1 menu item in the dropdown list
            // to prevent the list from being positioned in the upper-left corner
            // of the screen.
            DropDownItems.Add(_dummyItem);

            // Special menu items that should NOT be destroyed
            var specialItems = new ToolStripItem[] { _noDiscItem, _scanningItem, _dividerItem };

            // Disc Drive menu items
            var destroyableItems = AllMenuItems.Except(specialItems).Except(new[] { _dummyItem }).ToArray();

            foreach (var menuItem in destroyableItems)
            {
                DestroyMenuItem(menuItem);
                DropDownItems.Remove(menuItem);
            }

            foreach (var menuItem in specialItems)
            {
                DropDownItems.Remove(menuItem);
            }
        }

        private void UpdateMenu(ToolStripItem[] menuItems)
        {
            Logger.InfoFormat("Found {0} discs", menuItems.Length);

            var selectionState = new MenuSelectionState(AllMenuItems);

            ClearMenu();
            PopulateMenuSync(menuItems);

            selectionState.Restore(AllMenuItems);
        }

        private void PopulateMenuSync(ToolStripItem[] items)
        {
            DropDownItems.AddRange(items);

            var menuItems = AllMenuItems.Except(new[] { _dummyItem }).ToArray();

            if (!menuItems.Any())
            {
                DropDownItems.Add(_noDiscItem);
            }

            // We need to always keep at least 1 menu item in the dropdown list
            // to prevent the list from being positioned in the upper-left corner
            // of the screen.
            DropDownItems.Remove(_dummyItem);
        }

        #endregion

        #region Menu item creation / destruction

        private ToolStripItem[] CreateToolStripItems(DriveInfo[] drives)
        {
            return drives.Select(TryCreateMenuItem).Where(item => item != null).ToArray();
        }

        [CanBeNull]
        private ToolStripItem TryCreateMenuItem(DriveInfo driveInfo)
        {
            try { return CreateMenuItem(driveInfo); }
            catch (Exception e) { Logger.InfoFormat("Ignoring exception: {0}", e); }
            return null;
        }

        /// <exception cref="IOException">An I/O error occurred (for example, a disk error or a drive was not ready).</exception>
        /// <exception cref="DriveNotFoundException">The drive is not mapped or does not exist.</exception>
        /// <exception cref="SecurityException">The caller does not have the required permission.</exception>
        /// <exception cref="UnauthorizedAccessException">The volume label is being set on a network or CD-ROM drive.-or-Access to the drive information is denied.</exception>
        private ToolStripItem CreateMenuItem(DriveInfo driveInfo)
        {
            var driveLetter = driveInfo.Name;
            var volumeLabel = driveInfo.VolumeLabel;
            var text = driveLetter == volumeLabel ? volumeLabel : string.Format("{0} {1}", driveLetter, volumeLabel);
            var menuItem = new ToolStripMenuItem(text) { Tag = driveInfo };
            menuItem.Click += MenuItemOnClick;
            return menuItem;
        }

        private void DestroyMenuItem(ToolStripItem menuItem)
        {
            menuItem.Tag = null;
            menuItem.Click -= MenuItemOnClick;
        }

        #endregion

        #region Selection state management

        /// <summary>
        /// Captures and restores the selection state (i.e., which item is selected) of a dropdown menu.
        /// </summary>
        private class MenuSelectionState
        {
            private readonly int? _selectedIndex;
            private readonly DriveInfo _selectedDrive;

            /// <summary>
            /// Constructs a new <see cref="MenuSelectionState"/> object and captures the selection state
            /// (i.e., which menu item is selected, if any) from the given menu items.
            /// </summary>
            /// <param name="oldMenuItems">Items present in the dropdown menu <b>before</b> being repopulated.</param>
            public MenuSelectionState(ToolStripItem[] oldMenuItems)
            {
                var selectedItem = oldMenuItems.FirstOrDefault(item => item.Selected);
                if (selectedItem == null) return;

                _selectedIndex = oldMenuItems.ToList().IndexOf(selectedItem);
                _selectedDrive = selectedItem.Tag as DriveInfo;
            }

            /// <summary>
            /// Selects the same menu item that was selected previously.
            /// If the previously selected item is no longer in the list,
            /// the item at the same index is selected.
            /// </summary>
            /// <param name="newMenuItems">Items present in the dropdown menu <b>after</b> being repopulated.</param>
            public void Restore(ToolStripItem[] newMenuItems)
            {
                if (RestoreSelectedDrive(newMenuItems)) return;
                if (RestoreSelectedIndex(newMenuItems)) return;
            }

            private bool RestoreSelectedDrive(ToolStripItem[] newMenuItems)
            {
                if (_selectedDrive == null) return false;

                var itemToSelect = newMenuItems.FirstOrDefault(item => _selectedDrive.IsEqualTo(item));
                if (itemToSelect == null) return false;

                itemToSelect.Select();

                return true;
            }

            private bool RestoreSelectedIndex(ToolStripItem[] newMenuItems)
            {
                if (!_selectedIndex.HasValue) return false;
                if (!newMenuItems.Any()) return false;
                if (_selectedIndex < 0) return false;

                if (_selectedIndex < newMenuItems.Length)
                {
                    newMenuItems[_selectedIndex.Value].Select();
                }
                else
                {
                    newMenuItems.Last().Select();
                }

                return true;
            }
        }

        #endregion
    }

    internal static class DriveInfoExtensions
    {
        public static bool IsEqualTo(this DriveInfo thisDrive, ToolStripItem item)
        {
            if (item == null || thisDrive == null) return false;
            return thisDrive.IsEqualTo(item.Tag as DriveInfo);
        }

        private static bool IsEqualTo(this DriveInfo drive1, DriveInfo drive2)
        {
            if (drive1 == null || drive2 == null) return false;
            return string.Equals(drive1.Name, drive2.Name);
        }
    }

    public delegate void DiscMenuItemClickHandler(DriveInfo driveInfo);
}
