﻿// Copyright 2012-2014 Andrew C. Dvorak
//
// This file is part of BDHero.
//
// BDHero is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// BDHero is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with BDHero.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Windows.Forms;
using BDHero.JobQueue;
using BDHeroGUI.Forms;
using BDHeroGUI.Properties;
using DotNetUtils;
using DotNetUtils.Annotations;
using DotNetUtils.Concurrency;
using DotNetUtils.Extensions;
using UILib.WinForms.Controls;

namespace BDHeroGUI.Components
{
    public partial class MediaPanel : UserControl
    {
        private static readonly log4net.ILog Logger =
            log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        private const double DefaultRatio = 2.0 / 3.0;

        private readonly Hyperlink _hyperlink;

        #region Public getter/setter properties

        [CanBeNull]
        public ICoverArt SelectedCoverArt
        {
            get
            {
                return _selectedCoverArt;
            }
            private set
            {
                _selectedCoverArt = value;
                pictureBox.Image = SelectedCoverArt != null ? SelectedCoverArt.Image : Resources.no_poster_w185;
                AutoResize();
            }
        }

        private ICoverArt _selectedCoverArt;

        public Job Job
        {
            set
            {
                _job = value;
                LoadSearchResults();
            }
        }

        private Job _job;

        public bool SearchLinkEnabled
        {
            get { return linkLabelSearch.Enabled; }
            set { linkLabelSearch.Enabled = value; }
        }

        #endregion

        #region Private properties

        private ReleaseMedium SelectedReleaseMedium
        {
            get { return _job != null ? _job.SelectedReleaseMedium : null; }
        }

        private string SelectedUrl
        {
            get
            {
                var medium = SelectedReleaseMedium;
                return medium != null ? medium.Url : null;
            }
        }

        #endregion

        #region Public events

        public event EventHandler SelectedMediaChanged;

        public Action Search;

        #endregion

        public MediaPanel()
        {
            InitializeComponent();
            _hyperlink = new Hyperlink(pictureBox);
            Resize += (sender, args) => AutoResize();
            comboBoxMedia.SelectedIndexChanged += (sender, args) => OnSelectedMediaChanged();
            LoadSearchResults();
        }

        #region Private controller methods

        private void AutoResize()
        {
            var ratio = DefaultRatio;
            if (SelectedCoverArt != null && SelectedCoverArt.Image != null)
            {
                var image = SelectedCoverArt.Image;
                ratio = ((double) image.Width) / image.Height;
            }
            var width = ratio * pictureBox.Height;
            splitContainer.SplitterDistance = (int) Math.Round(width);
        }

        private void LoadSearchResults()
        {
            var media = _job != null
                            ? _job.Movies.OfType<ReleaseMedium>().ToList()
                            : new Collection<ReleaseMedium>().ToList();

            // TODO: Insert "No metadata" item
//            media.Insert(0, new Movie()
//                            {
//                                Title = "No metadata"
//                            });

            PopulateComboBox(media);
            OnSelectedMediaChanged();
        }

        private void PopulateComboBox(ICollection<ReleaseMedium> releaseMedia)
        {
            comboBoxMedia.Items.Clear();

            releaseMedia.ForEach(item => comboBoxMedia.Items.Add(item));

            var hasItems = releaseMedia.Any();

            if (hasItems)
                comboBoxMedia.SelectedIndex = 0;

            comboBoxMedia.Enabled = hasItems;
        }

        private void OnSelectedMediaChanged()
        {
            if (comboBoxMedia.Items.Count > 0)
                _job.Movies.ForEach(AutoSelect);

            LoadCoverArt();

            _hyperlink.Url = SelectedUrl;

            if (SelectedMediaChanged != null)
                SelectedMediaChanged(this, EventArgs.Empty);
        }

        private void AutoSelect(Movie movie, int i)
        {
            movie.IsSelected = (i == comboBoxMedia.SelectedIndex);
        }

        private void LoadCoverArt()
        {
            SelectedCoverArt = null;

            var medium = SelectedReleaseMedium;
            if (medium == null) return;

            var coverArt = medium.CoverArtImages.FirstOrDefault();
            if (coverArt == null) return;

            new EmptyPromise(this)
                .Work(delegate
                      {
                          var image = coverArt.Image;
                          Logger.DebugFormat("Finished loading poster image: {0}", image);
                      })
                .Fail(delegate(IPromise<Nil> promise)
                      {
                          Logger.Error("Unable to fetch poster image", promise.LastException);
                          SelectedCoverArt = null;
                      })
                .Done(delegate
                      {
                          SelectedCoverArt = coverArt;
                      })
                .Start()
                ;
        }

        #endregion

        #region UI event handlers

        private void linkLabelSearch_Click(object sender, EventArgs e)
        {
            if (Search != null)
                Search();
        }

        #endregion

        private void linkLabelCustom_Click(object sender, EventArgs e)
        {
            using (var form = new FormMediaCustom(SelectedReleaseMedium as Movie))
            {
                form.ShowDialog(this);
            }
        }
    }
}
